/*!
   @author Wu Shin-Ting - FEEC/Unicamp
   @date March/2016
   @see http://www.opengl-tutorial.org/beginners-tutorials/
 */

#include <iostream>

//#ifdef WIN32
//#define GLEW_STATIC
#include <GL/glew.h>
//#endif
//#define FREEGLUT_STATIC
#include <GL/freeglut.h>

#include <math.h>
#include <glm/glm.hpp> 
#include <glm/gtc/matrix_transform.hpp> 
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/type_ptr.hpp>  //! memory layout access
#include "LoadShaders.h"
#include <stdio.h>

#include "SOIL.h"

#ifndef WIN32
#include <time.h>
#endif 

//! Function scope: std 
using namespace std;

//! Function declaration (prototype)
void Initialize(const char * title);
void Display(void);
void Finalize(void);
void Reshape(int width, int height);
int checkError(void);

float aspect;

//! Data buffer
GLuint vao;
GLuint cube_vbo;
GLuint vertexbuffer;
GLuint colorbuffer;
GLuint normalbuffer;
GLuint texbuffer;

//! Texture ID
GLuint texture;

//! Variable locations in GPU
GLuint cube_prog;
GLint normal_mat_loc;
GLint mv_mat_loc;
GLint prj_mat_loc;
GLint tex_loc;

#ifndef WIN32
timespec start_tick;
#endif 

/*!
* OpenGL Error Codes
*/
int checkError()
{
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    if (glErr != GL_NO_ERROR)
    {
      std::cout << "glError = ";  

      switch (glErr) {
      case GL_INVALID_ENUM:
	std::cout << "GL_INVALID_NUM";
        break;
      case GL_INVALID_VALUE:
	std::cout << "GL_INVALID_VALUE";
        break;
      case GL_INVALID_OPERATION:
	std::cout << "GL_INVALID_OPERATION";
        break;
      case GL_INVALID_FRAMEBUFFER_OPERATION:
	std::cout << "GL_INVALID_FRAMEBUFFER_OPERATION";
        break;
      case GL_OUT_OF_MEMORY:
	std::cout << "GL_OUT_OF_MEMORY";
        break;
      case GL_STACK_UNDERFLOW:
	std::cout << "GL_STACK_UNDERFLOW";
        break;
      case GL_STACK_OVERFLOW:
	std::cout << "GL_STACK_OVERFLOW";
        break;
      }
      std::cout << std::endl;
      retCode = 1;
    }
    return retCode;
}

/*!
* Initialize OpenGL pipeline
*/
void Initialize()
{
  //! Create the program for rendering the scene from the light's POV. 
  glCreateProgram();

  ShaderInfo cube_info[] =
    {
      { GL_VERTEX_SHADER, "shaders/hello_cube.vs.glsl" },
      { GL_FRAGMENT_SHADER, "shaders/hello_cube.fs.glsl" },
      { GL_NONE, NULL }
    };

  cube_prog = LoadShaders(cube_info);

  //! Get the location in GPU
  mv_mat_loc = glGetUniformLocation(cube_prog, "MVMatrix");
  normal_mat_loc = glGetUniformLocation(cube_prog, "NormalMatrix");
  prj_mat_loc = glGetUniformLocation(cube_prog, "PMatrix");
  tex_loc = glGetUniformLocation(cube_prog, "Texture");

  static const GLfloat g_vertex_buffer_data[] = {
    -1.0f, 1.0f, 1.0f, //! triangle 1 : begin
    -1.0f,-1.0f, 1.0f,
    -1.0f,-1.0f,-1.0f, //! triangle 1 : end
    -1.0f, 1.0f,-1.0f, //! triangle 2 : begin
    1.0f, 1.0f,-1.0f, 
    -1.0f,-1.0f,-1.0f, //! triangle 2 : end
    -1.0f,-1.0f,-1.0f,
    1.0f,-1.0f, 1.0f,
    1.0f,-1.0f,-1.0f, //! triangle 3 : end
    1.0f,-1.0f,-1.0f,
    1.0f, 1.0f,-1.0f,
    -1.0f,-1.0f,-1.0f, //! triangle 4 : end
    -1.0f,-1.0f,-1.0f,
    -1.0f, 1.0f,-1.0f,
    -1.0f, 1.0f, 1.0f, //! triangle 5 : end
    -1.0f,-1.0f, 1.0f,
    1.0f,-1.0f, 1.0f,
    -1.0f,-1.0f,-1.0f, //! triangle 6 : end
    -1.0f,-1.0f, 1.0f,
    -1.0f, 1.0f, 1.0f,
    1.0f,-1.0f, 1.0f, //! triangle 7 : end
    1.0f,-1.0f,-1.0f,
    1.0f, 1.0f, 1.0f,
    1.0f, 1.0f,-1.0f, //! triangle 8 : end
    1.0f, 1.0f, 1.0f,
    1.0f,-1.0f,-1.0f,
    1.0f,-1.0f, 1.0f, //! triangle 9 : end
    1.0f, 1.0f,-1.0f,
    1.0f, 1.0f, 1.0f,
    -1.0f, 1.0f,-1.0f, //! triangle 10 : end
    -1.0f, 1.0f,-1.0f,
    1.0f, 1.0f, 1.0f,
    -1.0f, 1.0f, 1.0f, //! triangle 11 : end
    -1.0f, 1.0f, 1.0f,
    1.0f, 1.0f, 1.0f,
    1.0f,-1.0f, 1.0f //! triangle 12 : end
  };

  //! One color for each vertex. They were generated randomly.
  static const GLfloat g_color_buffer_data[] = {
    0.583f,  0.771f,  0.014f,
    0.609f,  0.115f,  0.436f,
    0.327f,  0.483f,  0.844f,
    0.822f,  0.569f,  0.201f,
    0.435f,  0.602f,  0.223f,
    0.310f,  0.747f,  0.185f,
    0.597f,  0.770f,  0.761f,
    0.559f,  0.436f,  0.730f,
    0.359f,  0.583f,  0.152f,
    0.483f,  0.596f,  0.789f,
    0.559f,  0.861f,  0.639f,
    0.195f,  0.548f,  0.859f,
    0.014f,  0.184f,  0.576f,
    0.771f,  0.328f,  0.970f,
    0.406f,  0.615f,  0.116f,
    0.676f,  0.977f,  0.133f,
    0.971f,  0.572f,  0.833f,
    0.140f,  0.616f,  0.489f,
    0.997f,  0.513f,  0.064f,
    0.945f,  0.719f,  0.592f,
    0.543f,  0.021f,  0.978f,
    0.279f,  0.317f,  0.505f,
    0.167f,  0.620f,  0.077f,
    0.347f,  0.857f,  0.137f,
    0.055f,  0.953f,  0.042f,
    0.714f,  0.505f,  0.345f,
    0.783f,  0.290f,  0.734f,
    0.722f,  0.645f,  0.174f,
    0.302f,  0.455f,  0.848f,
    0.225f,  0.587f,  0.040f,
    0.517f,  0.713f,  0.338f,
    0.053f,  0.959f,  0.120f,
    0.393f,  0.621f,  0.362f,
    0.673f,  0.211f,  0.457f,
    0.820f,  0.883f,  0.371f,
  };

  //! One normal for each vertex. 
  static const GLfloat g_normal_buffer_data[] = {
    -1.0f, 0.0f, 0.0f, //! triangle 1 : begin
    -1.0f, 0.0f, 0.0f,
    -1.0f, 0.0f, 0.0f, //! triangle 1 : end
    0.0f, 0.0f,-1.0f, //! triangle 2 : begin
    0.0f, 0.0f,-1.0f, 
    0.0f, 0.0f,-1.0f, //! triangle 2 : end
    0.0f,-1.0f,0.0f,
    0.0f,-1.0f,0.0f,
    0.0f,-1.0f,0.0f, //! triangle 3 : end
    0.0f, 0.0f,-1.0f,
    0.0f, 0.0f,-1.0f,
    0.0f, 0.0f,-1.0f, //! triangle 4 : end
    -1.0f, 0.0f, 0.0f,
    -1.0f, 0.0f, 0.0f,
    -1.0f, 0.0f, 0.0f, //! triangle 5 : end
    0.0f,-1.0f, 0.0f,
    0.0f,-1.0f, 0.0f,
    0.0f,-1.0f, 0.0f, //! triangle 6 : end
    0.0f, 0.0f, 1.0f,
    0.0f, 0.0f, 1.0f,
    0.0f, 0.0f, 1.0f, //! triangle 7 : end
    1.0f, 0.0f, 0.0f,
    1.0f, 0.0f, 0.0f,
    1.0f, 0.0f, 0.0f, //! triangle 8 : end
    1.0f, 0.0f, 0.0f,
    1.0f, 0.0f, 0.0f,
    1.0f, 0.0f, 0.0f, //! triangle 9 : end
    0.0f, 1.0f, 0.0f,
    0.0f, 1.0f, 0.0f,
    0.0f, 1.0f, 0.0f, //! triangle 10 : end
    0.0f, 1.0f, 0.0f,
    0.0f, 1.0f, 0.0f,
    0.0f, 1.0f, 0.0f, //! triangle 11 : end
    0.0f, 0.0f, 1.0f,
    0.0f, 0.0f, 1.0f,
    0.0f, 0.0f, 1.0f //! triangle 12 : end
  };

   //! One texture sample for each vertex. 
  static const GLfloat g_tex_buffer_data[] = {
    1.0f, 1.0f,        //! triangle 1 : begin
    1.0f, 0.0f, 
    0.0f, 0.0f,        //! triangle 1 : end
    1.0f, 1.0f,        //! triangle 2 : begin
    0.0f, 1.0f, 
    1.0f, 0.0f,        //! triangle 2 : end
    1.0f, 1.0f,
    0.0f, 0.0f,
    0.0f, 1.0f,        //! triangle 3 : end
    0.0f, 0.0f,
    0.0f, 1.0f,
    1.0f, 0.0f,        //! triangle 4 : end
    //   0.0f, 0.0f,        //! triangle 4 : end
    0.0f, 0.0f,
    0.0f, 1.0f,
    1.0f, 1.0f,        //! triangle 5 : end
    1.0f, 0.0f,
    0.0f, 0.0f,
    1.0f, 1.0f,        //! triangle 6 : end
    0.0f, 0.0f, 
    0.0f, 1.0f,
    1.0f, 0.0f,        //! triangle 7 : end
    1.0f, 0.0f,
    0.0f, 1.0f,
    1.0f, 1.0f,       //! triangle 8 : end
    0.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,       //! triangle 9 : end
    1.0f, 1.0f,
    1.0f, 0.0f, 
    0.0f, 1.0f,       //! triangle 10 : end
    0.0f, 1.0f,
    1.0f, 0.0f,
    0.0f, 0.0f,       //! triangle 11 : end
    0.0f, 1.0f, 
    1.0f, 1.0f,
    1.0f, 0.0f         //! triangle 12 : end
  };

  char buf[1024];
  glGetProgramInfoLog(cube_prog, 1024, NULL, buf);
#ifdef _DEBUG
  std::cout << "Program Info Log: " << buf << std::endl;
#endif

  //! Create a vertex array object
  glGenVertexArrays(1, &vao);
  glBindVertexArray(vao);
  glGenBuffers(1, &vertexbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(g_vertex_buffer_data), g_vertex_buffer_data, GL_STATIC_DRAW);
  //! Tell OpenGL the format of your position array data
  glVertexAttribPointer(
			0,                                // attribute. No particular reason for 0, but must match the layout in the shader.
			3,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);

  glGenBuffers(1, &colorbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(g_color_buffer_data), g_color_buffer_data, GL_STATIC_DRAW);
  //! Tell OpenGL the format of your color array data
  glVertexAttribPointer(
			1,                                // attribute. No particular reason for 1, but must match the layout in the shader.
			3,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);

  glGenBuffers(1, &normalbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(g_normal_buffer_data), g_normal_buffer_data, GL_STATIC_DRAW);
  //! Tell OpenGL the format of your normal array data
  glVertexAttribPointer(
			2,                                // attribute. No particular reason for 1, but must match the layout in the shader.
			3,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);

  glGenBuffers(1, &texbuffer);
  glBindBuffer(GL_ARRAY_BUFFER, texbuffer);
  glBufferData(GL_ARRAY_BUFFER, sizeof(g_tex_buffer_data), g_tex_buffer_data, GL_STATIC_DRAW);
  //! Tell OpenGL the format of your color array data
  glVertexAttribPointer(
			3,                                // attribute. No particular reason for 1, but must match the layout in the shader.
			2,                                // size
			GL_FLOAT,                         // type
			GL_FALSE,                         // normalized?
			0,                                // stride
			(void*)0                          // array buffer offset
			);


#ifndef WIN32
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start_tick);
#endif

  //! Load image for texturing (http://learnopengl.com/#!Getting-started/Textures)
  int width, height;
  unsigned char* image = SOIL_load_image("container.jpg", &width, &height, 0, SOIL_LOAD_RGB);
  glGenTextures(1, &texture);  
  glBindTexture(GL_TEXTURE_2D, texture);  
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, image);
  glGenerateMipmap(GL_TEXTURE_2D);
  SOIL_free_image_data(image);  //! free space
  glBindTexture(GL_TEXTURE_2D, 0); //! unbinding

  //! Clear buffers
  glClearColor(0.0f, 0.25f, 0.3f, 1.0f);
  glClearDepth(1.0f);

  //! Configure fixed functions
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glDisable(GL_CULL_FACE);
}

/*!
* Display event handling: Rendering
*/
void Display(void)
{
#ifdef WIN32
  static const unsigned int start_time = GetTickCount();
  float t = float((GetTickCount() - start_time)) / float(0x3FFF);
#else
  timespec tick;
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &tick);
  float t = float(tick.tv_nsec-start_tick.tv_nsec)/float(0x1FFFFFFF);
#endif

  static const glm::vec3 X(1.0f, 0.0f, 0.0f);
  static const glm::vec3 Y(0.0f, 1.0f, 0.0f);
  static const glm::vec3 Z(0.0f, 0.0f, 1.0f);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //! Activate shader program
  glUseProgram(cube_prog);
  glClear(GL_DEPTH_BUFFER_BIT);

  //! Modelview transformation 
  glm::mat4 tc_matrix =  glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, -4.0f)) * 
  glm::scale(glm::mat4(1.0f),glm::vec3(0.5f)) * glm::rotate(glm::mat4(1.0f), 5.0f*t, Y) *
  glm::rotate(glm::mat4(1.0f), 5.0f*t, Z) * glm::rotate(glm::mat4(1.0f), 5.0f*t, X);
  glUniformMatrix4fv(mv_mat_loc, 1, GL_FALSE, glm::value_ptr(tc_matrix));

  //! Print modelview matrix (column-major)
/*
  const float *pSource = (const float*)glm::value_ptr(tc_matrix);
    std::cout << "Modelview matrix" << std::endl;
    std::cout <<  pSource[0] << " " << pSource[4] << " " << pSource[8] << " " << pSource[12] << std::endl;
    std::cout <<  pSource[1] << " " << pSource[5] << " " << pSource[9] << " " << pSource[13] << std::endl;
    std::cout <<  pSource[2] << " " << pSource[6] << " " << pSource[10] << " " << pSource[14] << std::endl;
    std::cout <<  pSource[3] << " " << pSource[7] << " " << pSource[11] << " " << pSource[15] << std::endl;
*/

  //! Modelview inverse transformation
  glm::mat4 tc1_matrix = glm::transpose(glm::inverse(glm::scale(glm::mat4(1.0f),glm::vec3(0.5f)) * glm::rotate(glm::mat4(1.0f), 5.0f*t, Y) *
  glm::rotate(glm::mat4(1.0f), 5.0f*t, Z) * glm::rotate(glm::mat4(1.0f), 5.0f*t, X)));
  glUniformMatrix4fv(normal_mat_loc, 1, GL_FALSE, glm::value_ptr(tc1_matrix));
/*
    std::cout << "Transpose inverse modelview matrix" << std::endl;
    pSource = (const float*)glm::value_ptr(tc1_matrix);
    std::cout <<  pSource[0] << " " << pSource[4] << " " << pSource[8] << " " << pSource[12] << std::endl;
    std::cout <<  pSource[1] << " " << pSource[5] << " " << pSource[9] << " " << pSource[13] << std::endl;
    std::cout <<  pSource[2] << " " << pSource[6] << " " << pSource[10] << " " << pSource[14] << std::endl;
    std::cout <<  pSource[3] << " " << pSource[7] << " " << pSource[11] << " " << pSource[15] << std::endl;
*/

  //! Perspective projection matrix
  float aux = (M_PI*30.0f)/180.f;
  tc_matrix = glm::perspective(aux, 1.0f/aspect, 0.1f, 10.0f);
  glUniformMatrix4fv(prj_mat_loc, 1, GL_FALSE, glm::value_ptr(tc_matrix));

  //! Set ligting parameters
  GLfloat tmp[3];
  tmp[0] = tmp[1] = tmp[2] = 0.50;
  glUniform3fv (glGetUniformLocation(cube_prog, "Ambient"), 1, tmp);
  tmp[0] = tmp[1] = tmp[2] = 1.0;
  glUniform3fv (glGetUniformLocation(cube_prog, "LightColor"), 1, tmp);
  tmp[0] = 0.0;  tmp[1] = 1.; tmp[2] = 0.;
  glUniform3fv (glGetUniformLocation(cube_prog, "LightDirection"), 1, tmp);
  tmp[0] = 0.0; tmp[1] = 1./sqrt(2.); tmp[2] = 1./sqrt(2.); 
  glUniform3fv (glGetUniformLocation(cube_prog, "HalfVector"), 1, tmp);
  tmp[0] = 0.0; tmp[1] = 0.0; tmp[2] = 1.; 
  glUniform3fv (glGetUniformLocation(cube_prog, "ViewVector"), 1, tmp);
  glUniform1f(glGetUniformLocation(cube_prog, "Shininess"), 2.5);
  glUniform1f(glGetUniformLocation(cube_prog, "Strength"), 1.);
  //! According to Fig. 7.19 of Real-Time Rendering Rf(0)=0.3 for glass 
  glUniform1f(glGetUniformLocation(cube_prog, "F0"), 0.3);

  //! Bind texture 
  glUniform1i(tex_loc, 0);
  glActiveTexture(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D, texture);
  
  //! Enable the vertex attributes for rendering
  glBindVertexArray(vao);
  glEnableVertexAttribArray(0);
  glEnableVertexAttribArray(1);
  glEnableVertexAttribArray(2);
  glEnableVertexAttribArray(3);
  glDrawArrays(GL_TRIANGLES, 0, 12*3);  

  //! Swap front and back buffers
  glutSwapBuffers();
  //! Force redisplay
  glutPostRedisplay();
}

/*!
* Keyboard event handling
*/
void Keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case 27:
    glUseProgram(0);
    glDeleteProgram(cube_prog);
    exit(0);
    break;
  default:
    break;
  }
}

/*!
* Window resize event handling: Resizing the display (and generating display event) 
*/
void Reshape(int width, int height)
{
  glViewport(0, 0 , width, height);

  aspect = float(height) / float(width);
}

int main(int argc, char ** argv)                            
{                                                           
#ifdef _DEBUG
  glutInitContextFlags(GLUT_DEBUG);
#endif

  //! Connect OpenGL with a Window system
  glutInit(&argc, argv);
  glutInitContextProfile(GLUT_CORE_PROFILE);
  glutInitContextVersion(3, 1);
  glutInitWindowSize(1024, 768);
  glutInitWindowPosition (140, 140);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  glutCreateWindow("OpenGL Application");

  //! Specify callback functions
  glutDisplayFunc(Display);
  glutReshapeFunc(Reshape);
  glutKeyboardFunc(Keyboard);
  //glutIdleFunc(Display);
 
  //! Initialize glew
  //#ifdef WIN32
  glewExperimental = GL_TRUE;
  glewInit();
  std::cout << "OpenGL version supported by this platform (" << glGetString(GL_VERSION) << ")" << std::endl;
  //#endif

#if defined _DEBUG && defined WIN32
  if (glDebugMessageCallbackARB != NULL)
    glDebugMessageCallbackARB(DebugOutputCallback, this);
#endif

  //! Initialize Opengl context
  Initialize(); 
                                
  //! Endless loop waiting for events
  glutMainLoop();                               
                                                            
  return 0;                                               
}                                                           
 
